#region Using Statements
using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Storage;
#endregion

namespace KeyboardInstrument {

	public class Audio : GameComponent {
		readonly AudioEngine audioEngine;
		readonly SoundBank soundBank;
		readonly WaveBank waveBank;

		public override void Update(GameTime gameTime) {
			audioEngine.Update();
			base.Update(gameTime);
		}

		public Cue this[string cueName] {
			get {
				try {
					return soundBank.GetCue(cueName);
				} catch {
					return null;
				}
			}
		}

		public Audio(Game game)
			: base(game) {
			audioEngine = new AudioEngine("content/audio/KeyboardInstrument.xgs");
			soundBank = new SoundBank(audioEngine, "content/audio/Sound Bank.xsb");
			waveBank = new WaveBank(audioEngine, "content/audio/Wave Bank.xwb");
		}
	}
	public class InstrumentKey : GameComponent {
		readonly Vector2 position;
		readonly Rectangle sourceRectangle;
		readonly string cueName;
		readonly Instrument instrument;
		bool pressed;
		float volume;
		Cue cue;

		public InstrumentKey(Game game, Instrument instrument, string cueName, Vector2 position, Rectangle sourceRectangle)
			: base(game) {
			this.instrument = instrument;
			this.cueName = cueName;
			this.position = position;
			this.sourceRectangle = sourceRectangle;
		}
		public Cue Cue { get { return cue; } }
		public bool Velocity { get { return pressed; } }
		public Vector2 Position { get { return position; } }
		public Rectangle SourceRectangle { get { return sourceRectangle; } }

		public void Press() {
			if (!pressed) {
				pressed = true;
				volume = 1;
				cue = instrument.Audio[cueName];
				if (cue == null) {
					return;
				}
				try {
					cue.Play();
				} catch {
				}
			}
		}
		public void Release() {
			pressed = false;
		}

		public override void Update(GameTime gameTime) {
			if (!pressed) {
				volume -= 0.01f;
			} else {
				volume -= 0.006f;
			}
			if (volume < 0) {
				volume = 0;
				if (cue != null) {
					cue.Stop(AudioStopOptions.AsAuthored);
					cue = null;
				}
			}
			if (cue != null) {
				cue.SetVariable("Volume", volume);
			}
			base.Update(gameTime);
		}
	}
	public class Instrument : GameComponent {
		public const int KeysPerOctave = 12;
		const int MaximumOctave = 8;
		const int WidthPerOctave = 0x2a0;
		public int MinimumKey { get { return 0; } }
		public int MaximumKey { get { return keys.Length; } }

		readonly Audio audio;
		readonly InstrumentKey[] keys = new InstrumentKey[KeysPerOctave * MaximumOctave];

		public Instrument(Game game, Audio audio)
			: base(game) {
			this.audio = audio;
			for (int i = 0; i < keys.Length; i++) {
				int scale = i % KeysPerOctave;
				int kind = kinds[scale];
				int octave = i / KeysPerOctave;
				Vector2 position = new Vector2(octave * WidthPerOctave, 0);
				position.X += locations[scale].X;
				Rectangle sourceRectangle = keyRectangles[kind];
				game.Components.Add(keys[i] = new InstrumentKey(game, this, string.Format("{0}", i), position, sourceRectangle));
			}
		}

		public Audio Audio { get { return audio; } }
		public int HitTest(float x, float y, float scale) {
			Point location;
			location.X = (int)(x / scale);
			location.Y = (int)(y / scale);
			int octave = (int)Math.Floor((float)location.X / WidthPerOctave);

			if (octave < 0 || octave >= MaximumOctave)
				return -1;
			if (location.Y < 0 || location.Y >= 32 * 16)
				return -1;

			location.X %= WidthPerOctave;

			int index = location.X / 32;

			if (location.Y >= 32 * 11)
				index += 21;
			return keyMap[index] + octave * KeysPerOctave;

		}

		static readonly int[] keyMap = {
			00, 00, 01, 01, 02, 03, 03, 04, 04, 05, 05, 06, 06, 07, 08, 08, 09, 10, 10, 11, 11,
			00, 00, 00, 02, 02, 02, 04, 04, 04, 05, 05, 05, 07, 07, 07, 09, 09, 09, 11, 11, 11,
		};

		static readonly int[] kinds = { 0, 3, 1, 3, 2, 0, 3, 1, 3, 1, 3, 2 };
		static readonly Point[] locations = new Point[] {
			new Point(0x000, 0),
			new Point(0x040, 0),
			new Point(0x060, 0),
			new Point(0x0a0, 0),
			new Point(0x0c0, 0),
			new Point(0x120, 0),
			new Point(0x160, 0),
			new Point(0x180, 0),
			new Point(0x1c0, 0),
			new Point(0x1e0, 0),
			new Point(0x220, 0),
			new Point(0x240, 0),
		};

		static readonly Rectangle[] keyRectangles = new Rectangle[] {
			new Rectangle(32 * 00, 0, 32 * 3, 512),
			new Rectangle(32 * 04, 0, 32 * 3, 512),
			new Rectangle(32 * 08, 0, 32 * 3, 512),
			new Rectangle(32 * 12, 0, 32 * 2, 512),
		};

		public InstrumentKey this[int key] {
			get { return keys[key]; }
		}

		public void KeyOn(int key) {
			if (key < 0 || key >= keys.Length)
				return;
			keys[key].Press();
		}
		public void KeyOff(int key) {
			if (key < 0 || key >= keys.Length)
				return;
			keys[key].Release();
		}

	}

	public class KeyboardConsole : DrawableGameComponent {
		readonly ContentManager content;
		readonly Instrument instrument;
		SpriteBatch sprites;
		Texture2D texture;
		int hoveringKey;
		KeyboardState keyboardState;
		MouseState mouseState;
		Vector2 positionOffset = new Vector2(-300, 200);
		float scale = 1.0f / 4;

		public KeyboardConsole(Game game, ContentManager content, Instrument instrument)
			: base(game) {
			this.content = content;
			this.instrument = instrument;
		}

		protected override void LoadGraphicsContent(bool loadAllContent) {
			if (loadAllContent) {
				sprites = new SpriteBatch(GraphicsDevice);
				texture = content.Load<Texture2D>("content/texture/KeyboardInstrument");
			}
			base.LoadGraphicsContent(loadAllContent);
		}

		static readonly KeyValuePair<Keys, int>[] assigns = {
			new KeyValuePair<Keys,int>(Keys.A, 44),
			new KeyValuePair<Keys,int>(Keys.Z, 45),
			new KeyValuePair<Keys,int>(Keys.S, 46),
			new KeyValuePair<Keys,int>(Keys.X, 47),
			new KeyValuePair<Keys,int>(Keys.C, 48),
			new KeyValuePair<Keys,int>(Keys.F, 49),
			new KeyValuePair<Keys,int>(Keys.V, 50),
			new KeyValuePair<Keys,int>(Keys.G, 51),
			new KeyValuePair<Keys,int>(Keys.B, 52),
			new KeyValuePair<Keys,int>(Keys.N, 53),
			new KeyValuePair<Keys,int>(Keys.J, 54),
			new KeyValuePair<Keys,int>(Keys.M, 55),
			new KeyValuePair<Keys,int>(Keys.K, 56),
			new KeyValuePair<Keys,int>(Keys.OemComma, 57),
			new KeyValuePair<Keys,int>(Keys.L, 58),
			new KeyValuePair<Keys,int>(Keys.OemPeriod, 59),
			new KeyValuePair<Keys,int>(Keys.OemQuestion, 60),
		};

		public override void Update(GameTime gameTime) {
			KeyboardState previousKeyboardState = keyboardState;
			keyboardState = Keyboard.GetState();
			MouseState previousMouseState = mouseState;
			mouseState = Mouse.GetState();

			for (int i = 0; i < assigns.Length; i++) {
				KeyValuePair<Keys, int> assign = assigns[i];
				if (keyboardState.IsKeyDown(assign.Key)) {
					instrument.KeyOn(assign.Value);
				} else if (previousKeyboardState.IsKeyDown(assign.Key)) {
					instrument.KeyOff(assign.Value);
				}
			}

			int previousKey = hoveringKey;
			hoveringKey = instrument.HitTest(mouseState.X - positionOffset.X, mouseState.Y - positionOffset.Y, scale);
			if (mouseState.LeftButton == ButtonState.Pressed) {
				instrument.KeyOn(hoveringKey);
			} else if (previousMouseState.LeftButton == ButtonState.Pressed) {
				instrument.KeyOff(previousKey);
			}
			if (previousKey != hoveringKey) {
				instrument.KeyOff(previousKey);
			}
			base.Update(gameTime);
		}
		static readonly Color[] colors = {
			Color.White,
			Color.Red,
		};

		public Vector2 PositionOffset { get { return positionOffset; } set { positionOffset = value; } }
		public float Scale { get { return scale; } set { scale = value; } }

		public override void Draw(GameTime gameTime) {
			sprites.Begin();

			for (int i = instrument.MinimumKey; i < instrument.MaximumKey; i++) {
				InstrumentKey key = instrument[i];

				Vector2 position = key.Position * scale + positionOffset;

				sprites.Draw(texture, position, key.SourceRectangle,
					colors[key.Velocity ? 1 : 0], 0, Vector2.Zero, scale, SpriteEffects.None, 1);
			}
			sprites.End();
			base.Draw(gameTime);
		}
	}

	public class KeyboardInstrumentGame : Microsoft.Xna.Framework.Game {
		readonly GraphicsDeviceManager graphics;
		readonly ContentManager content;
		readonly Audio audio;
		readonly Instrument instrument;
		public KeyboardInstrumentGame() {
			graphics = new GraphicsDeviceManager(this);
			content = new ContentManager(Services);
			Components.Add(audio = new Audio(this));
			Components.Add(instrument = new Instrument(this, audio));
			Components.Add(new KeyboardConsole(this, content, instrument));
			IsMouseVisible = true;
		}

		protected override void UnloadGraphicsContent(bool unloadAllContent) {
			if (unloadAllContent) {
				content.Unload();
			}
		}

		protected override void Update(GameTime gameTime) {
			if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
				this.Exit();
			base.Update(gameTime);
		}

		protected override void Draw(GameTime gameTime) {
			graphics.GraphicsDevice.Clear(Color.RosyBrown);
			base.Draw(gameTime);
		}
	}

}
